#!/usr/bin/env python
# =============================================================================
# Pimmi CLI Endpoint
# =============================================================================
#
# CLI enpoint of the Pimmi library.
#
import os
import csv
import sys
import click
import logging

from pimmi import load_index, save_index, fill_index_mt, create_index_mt, get_index_images, query_index_mt
import pimmi.pimmi_parameters as prm
import pimmi.toolbox as tbx

CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"]
}

logger = logging.getLogger(__name__)


@click.group(context_settings=CONTEXT_SETTINGS)
def main():
    pass


@main.command(help="Create index and fill with vectors of images. Receive IMAGE-DIR, a directory containing images. "
                   "Index these images and save. INDEX-NAME will be used as index id by other pimmi commands.")
@click.argument("image-dir", type=click.Path(exists=True))
@click.argument("index-name", type=str)
@click.option("--index-path", type=str, help="Directory where the index should be stored/loaded from. "
                                             "Defaults to './index'", default="./index")
@click.option("--index-type", type=str, default="IVF1024,Flat")
@click.option("--load-faiss", is_flag=True, type=bool)
def fill(image_dir, index_name, index_path, index_type, load_faiss):

    if not os.path.isdir(image_dir):
        logger.error("The provided image-dir is not a directory.")
        sys.exit(1)

    if not index_name:
        index_name = os.path.basename(os.path.normpath(image_dir))

    logger.info("listing images recursively from : " + image_dir)
    images = tbx.get_all_images([image_dir])

    if load_faiss:
        previous_faiss_index = os.path.join(index_path, ".".join([index_name, index_type, "faiss"]))
        previous_faiss_meta = os.path.join(index_path, ".".join([index_name, index_type, "meta"]))
        previous_index = load_index(previous_faiss_index, previous_faiss_meta)
        logger.info("using " + str(prm.nb_images_to_train_index) + " images to fill index")
        filled_index = fill_index_mt(previous_index, images, image_dir)
    else:
        logger.info("using " + str(prm.nb_images_to_train_index) + " images to train and fill index")
        filled_index = create_index_mt(index_type, images, image_dir)

    if not os.path.isdir(index_path):
        if click.confirm("Are you sure you want to save index data in {}?".format(os.path.abspath(index_path))):
            os.mkdir(index_path)
        else:
            index_path = click.prompt('Please enter a valid directory', type=click.Path(exists=True))
            if not os.path.isdir(index_path):
                logger.error("{} does not exist.".format(os.path.abspath(index_path)))
                sys.exit(1)

    filled_faiss_index = os.path.join(index_path, ".".join([index_name, index_type, "faiss"]))
    filled_faiss_meta = os.path.join(index_path, ".".join([index_name, index_type, "meta"]))
    save_index(filled_index, filled_faiss_index, filled_faiss_meta)


@main.command(help="Query some index.")
@click.argument("index-name", type=str)
@click.argument("image-dir", type=click.Path(exists=True))
@click.option("--index-path", type=str, help="Directory where the index should be loaded from. "
                                             "Defaults to './index'", default="./index")
@click.option("--index-type", type=str, default="IVF1024,Flat")
@click.option("--nb-img", type=int, help="Nb images to query the index. Defaults to the total nb of images.")
@click.option('--nb_per_split', type=int, help="nb images to query per pack", default=10000)
@click.option('--simple', help="use only very simple query params", is_flag=True, type=bool)
def query(index_name, image_dir, index_path, index_type, nb_img, nb_per_split, simple):
    faiss_index = os.path.join(index_path, ".".join([index_name, index_type, "faiss"]))
    faiss_meta = os.path.join(index_path, ".".join([index_name, index_type, "meta"]))
    index = load_index(faiss_index, faiss_meta)

    images = get_index_images(index, image_dir)

    if nb_img:
        images = images.head(nb_img)

    logger.info("total number of queries " + str(len(images)))
    images = images.sort_values(by=prm.dff_image_path)
    queries = tbx.split_pack(images, nb_per_split)
    for pack in queries:
        pack_result_file = faiss_index.replace("faiss", "mining") + "_" + str(pack[prm.dff_pack_id]).zfill(6) + ".csv"
        logger.info("query " + str(len(pack[prm.dff_pack_files])) + " files from pack " +
                    str(pack[prm.dff_pack_id]) + " -> " + pack_result_file)
        query_result = query_index_mt(index, pack[prm.dff_pack_files], image_dir, pack=pack[prm.dff_pack_id])
        query_result = query_result.sort_values(by=[prm.dff_query_path, prm.dff_nb_match_ransac, prm.dff_ransac_ratio],
                                                ascending=False)
        query_result.to_csv(pack_result_file, index=False, quoting=csv.QUOTE_NONNUMERIC)



import logging
import pimmi
import pimmi_parameters as prm
import pimmi_toolbox as tbx
import argparse

logger = logging.getLogger("index")

index_type = "IVF1024,Flat"


parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
parser.add_argument("--action", required=True, choices=["create_empty", "fill", "info", "correct"])
parser.add_argument('--index', required=False, help="faiss factory string, default : " + index_type)
parser.add_argument('--thread', required=False, type=int, help="nb threads, default : " + str(prm.nb_threads))
parser.add_argument('--nb_img', required=False, type=int, help="nb images to train index (default : " +
                                                               str(prm.nb_images_to_train_index) +
                                                               ") or to add to index")
parser.add_argument('--load_faiss', required=False, help="faiss index file to load")
parser.add_argument('--save_faiss', required=False, help="faiss index file to save")
image_source = parser.add_mutually_exclusive_group()
image_source.add_argument('--images_dir', required=False, help="index all images from this directory")
image_source.add_argument('--images_meta', required=False, help="index all images from this file")
parser.add_argument('--images_meta_root', required=False, help="root path for meta relative paths")
args = parser.parse_args()

if __name__ == '__main__':
    if args.thread:
        prm.nb_threads = args.thread
    if args.nb_img:
        prm.nb_images_to_train_index = args.nb_img
    if args.index:
        index_type = args.index

    logger.info("action executed : " + args.action)

    if "info" == args.action:
        if args.load_faiss:
            previous_faiss_index = args.load_faiss + ".faiss"
            previous_faiss_meta = args.load_faiss + ".meta"
            previous_index = pimmi.load_index(previous_faiss_index, previous_faiss_meta)
        exit(0)

    if "correct" == args.action:
        if args.load_faiss:
            previous_faiss_index = args.load_faiss + ".faiss"
            previous_faiss_meta = args.load_faiss + ".meta"
            previous_index = pimmi.load_index(previous_faiss_index, previous_faiss_meta, correct=index_type)
            pimmi.save_index(previous_index, previous_faiss_index, previous_faiss_meta)
        exit(0)

    if not args.save_faiss:
        logger.error("save_faiss should be provided")
        exit(1)

    logger.info("using " + str(prm.nb_threads) + " threads")

    images = None
    images_root = None
    if args.images_dir:
        logger.info("listing images recursively from : " + args.images_dir)
        images = tbx.get_all_images([args.images_dir])
        images_root = args.images_dir
    if args.images_meta:
        if not args.images_meta_root:
            logger.error("images_meta_root should be provided")
            exit(1)
        logger.info("getting images from : " + args.images_meta)
        images = tbx.get_all_retrieved_images(args.images_meta, args.images_meta_root)
        images_root = args.images_meta_root
    if images is None:
        logger.error("you should provide a way to find images")
        exit(1)

    if args.nb_img:
        images = images[0:args.nb_img]

    if "create_empty" == args.action:
        if args.load_faiss:
            logger.error("unable to load an index before creating an empty one")
            exit(1)
        logger.info("index type : " + index_type)
        logger.info("using " + str(prm.nb_images_to_train_index) + " images to train index")
        empty_index = pimmi.create_index_mt(index_type, images, images_root, only_empty_index=True)
        empty_faiss_index = args.save_faiss + ".faiss"
        empty_faiss_meta = args.save_faiss + ".meta"
        pimmi.save_index(empty_index, empty_faiss_index, empty_faiss_meta)

    if "fill" == args.action:
        if args.load_faiss:
            previous_faiss_index = args.load_faiss + ".faiss"
            previous_faiss_meta = args.load_faiss + ".meta"
            previous_index = pimmi.load_index(previous_faiss_index, previous_faiss_meta)
            logger.info("using " + str(prm.nb_images_to_train_index) + " images to fill index")
            filled_index = pimmi.fill_index_mt(previous_index, images, images_root)
        else:
            logger.info("using " + str(prm.nb_images_to_train_index) + " images to train and fill index")
            filled_index = pimmi.create_index_mt(index_type, images, images_root)
        filled_faiss_index = args.save_faiss + ".faiss"
        filled_faiss_meta = args.save_faiss + ".meta"
        pimmi.save_index(filled_index, filled_faiss_index, filled_faiss_meta)



import logging
import csv
import pimmi
import pimmi_parameters as prm
import pimmi_toolbox as tbx
import argparse

nb_per_split = 10000

logger = logging.getLogger("query")

parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
parser.add_argument('--thread', required=False, type=int, help="nb threads, default : " + str(prm.nb_threads))
parser.add_argument('--nb_img', required=False, type=int, help="nb images to query the index")
parser.add_argument('--nb_per_split', required=False, type=int, help="nb images to query per pack")
parser.add_argument('--load_faiss', required=True, help="faiss index file to load")
parser.add_argument('--save_mining', required=True, help="mining file to save")
parser.add_argument('--simple', required=False, action='store_true', help="use only very simple query params")
image_source = parser.add_mutually_exclusive_group()
image_source.add_argument('--images_dir', required=False, help="query all images from this directory")
image_source.add_argument('--images_meta', required=False, help="query all images from this file")
image_source.add_argument('--images_mining', required=False, action='store_true', help="query all images from the index")
parser.add_argument('--images_root', required=False, help="root path for imagers relative paths")
args = parser.parse_args()


if __name__ == '__main__':
    if args.simple:
        # prm.nb_threads = 1
        prm.do_filter_on_sift_dist = False
        prm.adaptative_sift_nn = False

    if args.thread:
        prm.nb_threads = args.thread
    if args.nb_per_split:
        nb_per_split = args.nb_per_split

    prm.remove_query_from_results = False

    images = None
    images_root = None
    if not args.images_root:
        if args.images_meta or args.images_mining:
            logger.error("images_root should be provided")
            exit(1)

    if args.images_dir:
        logger.info("listing images recursively from : " + args.images_dir)
        # TODO: change to (id, path) df
        images = tbx.get_all_images([args.images_dir])
        images_root = args.images_dir
    if args.images_meta:
        logger.info("getting images from : " + args.images_meta)
        # TODO: change to (id, path) df
        images = tbx.get_all_retrieved_images(args.images_meta, args.images_root)
        images_root = args.images_root
    if args.images_mining:
        logger.info("querying the indexed dataset on himself")
        images = "loaded later"
        images_root = args.images_root
    if images is None:
        logger.error("you should provide a way to find images")
        exit(1)

    faiss_index = args.load_faiss + ".faiss"
    faiss_meta = args.load_faiss + ".meta"
    index = pimmi.load_index(faiss_index, faiss_meta)

    if args.images_mining:
        images = pimmi.get_index_images(index, images_root)

    if args.nb_img:
        images = images.head(args.nb_img)

    logger.info("total number of queries " + str(len(images)))

    images = images.sort_values(by=prm.dff_image_path)
    queries = tbx.split_pack(images, nb_per_split)
    for pack in queries:
        pack_result_file = args.save_mining + "_" + str(pack[prm.dff_pack_id]).zfill(6) + ".csv"
        logger.info("query " + str(len(pack[prm.dff_pack_files])) + " files from pack " +
                    str(pack[prm.dff_pack_id]) + " -> " + pack_result_file)
        query_result = pimmi.query_index_mt(index, pack[prm.dff_pack_files], images_root, pack=pack[prm.dff_pack_id])
        query_result = query_result.sort_values(by=[prm.dff_query_path, prm.dff_nb_match_ransac, prm.dff_ransac_ratio],
                                                ascending=False)
        query_result.to_csv(pack_result_file, index=False, quoting=csv.QUOTE_NONNUMERIC)



import pimmi_parameters as prm
import logging
import pickle
import pandas as pd
import numpy as np
import igraph as ig
import json
import pimmi

# TODO parameters from command line
logger = logging.getLogger("gener")

merged_meta_file = "index/dataset1.ivf1024.meta"
all_results_file = "index/dataset1.ivf1024.mining.all"
viz_data_file = "index/dataset1.ivf1024.mining.clusters.json"


def from_cluster_2col_table_to_list(data, keycol, valcol):
    keys, values = data.sort_values(keycol).values.T
    ukeys, index = np.unique(keys, True)
    arrays = np.split(values, index[1:])
    return pd.DataFrame({keycol: ukeys, valcol: [list(a) for a in arrays]})


if __name__ == '__main__':
    pd.set_option('display.max_rows', 500)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)
    pd.set_option('display.max_colwidth', 1000)

    logger.info("Loading results from " + all_results_file)
    with open(all_results_file, 'rb') as f:
        all_similarities = pickle.load(f)
    f.close()
    logger.info(" ~ similarity file : " + str(len(all_similarities)) + " edges")
    # print(all_similarities.head(5))

    with open(merged_meta_file, 'rb') as f:
        all_meta = pickle.load(f)
    f.close()
    all_meta = pimmi.meta_as_df(all_meta)
    logger.info("all_meta : %d", len(all_meta))

    # all_meta = all_meta.sort_values(by=[prm.dff_nb_points], ascending=True)
    # TODO check comment below
    # all_meta = all_meta[~pd.isna(all_meta[prm.dff_nb_points])]
    logger.info("all_meta filtered : %d", len(all_meta))
    # print(all_meta.head(5))

    all_similarities = all_similarities[["query_image_id", "result_image_id", "nb_match_ransac"]]

    all_results_with_meta = pd.merge(all_similarities, all_meta, how="left",
                                     left_on="query_image_id", right_on="image_id")
    all_results_with_meta = pd.merge(all_results_with_meta, all_meta, how="left", left_on="result_image_id",
                                     right_on="image_id", suffixes=("_query", "_result"))
    all_results_with_meta["query_ransac_ratio"] = all_results_with_meta["nb_match_ransac"] / \
                                                  all_results_with_meta["nb_points_query"]
    all_results_with_meta["result_ransac_ratio"] = all_results_with_meta["nb_match_ransac"] / \
                                                   all_results_with_meta["nb_points_result"]
    logger.info("all_results_with_meta : %d edges", len(all_results_with_meta))

    all_results_with_meta_graph = all_results_with_meta.query('nb_match_ransac >= 10')
    # all_results_with_meta_graph = all_results_with_meta.query('nb_match_ransac >= 10 and (query_ransac_ratio > 0.8 or result_ransac_ratio > 0.8)')
    # all_results_with_meta_graph = all_results_with_meta.query('nb_match_ransac >= 300 and (query_ransac_ratio > 0.5 or result_ransac_ratio > 0.5)')
    logger.info("all_results_with_meta_graph : %d edges", len(all_results_with_meta_graph))
    # print(all_results_with_meta_graph.head(100))

    g = ig.Graph.TupleList([(row["query_image_id"], row["result_image_id"], row["nb_match_ransac"])
                            for index, row in all_results_with_meta_graph.iterrows()], directed=True, weights=True)

    logger.info("Number of vertices in the graph: %d", g.vcount())
    logger.info("Number of edges in the graph: %d", g.ecount())
    logger.info("Is the graph directed: %d", g.is_directed())
    logger.info("Maximum degree in the graph: %d", g.maxdegree())


    comp = g.components(mode="weak")
    logger.info("Connected components in the graph: %d", len(comp))
    clusters = pd.DataFrame(list(zip(g.vs["name"], comp.membership)), columns=['image_id', 'cluster'])
    logger.info("Images in clusters : %d", len(clusters))

    clusters_with_meta = pd.merge(clusters, all_meta, how="left", left_on="image_id", right_on="image_id")
    # print(clusters_with_meta.head(10))

    # summary_clusters = clusters_with_meta.groupby("cluster").agg({prm.mex_nbSeen: ['sum', 'count'],
    #                                                               prm.mex_relative_path: ['first']})
    # summary_clusters.columns = ['nb_seen', 'nb_images', 'sample_path']
    # summary_clusters = summary_clusters.reset_index()
    # logger.info("%d distinct clusters", len(summary_clusters))
    # summary_clusters = summary_clusters.sort_values(by=['nb_seen', 'nb_images'], ascending=False)

    summary_clusters = clusters_with_meta.groupby("cluster").agg({prm.dff_image_path: ['first', 'count']})
    summary_clusters.columns = ['sample_path', 'nb_images']
    summary_clusters = summary_clusters.reset_index()
    logger.info("%d distinct clusters", len(summary_clusters))

    sub_graphs = comp.subgraphs()
    for cluster_id, sg in enumerate(sub_graphs):
        # cluster_id = 1
        # sg = sub_graphs[cluster_id]
        nb_points = sorted(clusters_with_meta[clusters_with_meta["cluster"] == cluster_id]["nb_points"])
        sum_weight = list(range(1, len(nb_points) + 1))
        sum_weight.reverse()
        max_theoretical_matches = 2 * sum([i * j for i, j in zip(nb_points, sum_weight)])
        nb_matches = sum(sg.es["weight"])
        quality = nb_matches / max_theoretical_matches
        summary_clusters.loc[cluster_id, "match_quality"] = quality
        logger.info("cluster %d has %d vertices and %d edges, quality index : %f", cluster_id, sg.vcount(), sg.ecount(), quality)

    print(summary_clusters.head(10))

    clusters_viz = from_cluster_2col_table_to_list(clusters_with_meta[["cluster", prm.dff_image_path]], "cluster", "images")
    clusters_viz = pd.merge(clusters_viz, summary_clusters, how="left", left_on="cluster", right_on="cluster")
    clusters_viz = clusters_viz.sort_values(by=['nb_images'], ascending=False)
    print(clusters_viz.head(10))

    # viz_data = list(clusters_viz["relativePath"])
    # # print(viz_data[0:5])
    # with open(viz_data_file, 'w') as f:
    #     json.dump(viz_data, f)
    # f.close()

    clusters_viz.to_json(viz_data_file, orient="records")



import pandas as pd
import csv
import glob
import pickle
import logging
import pimmi_parameters as prm
import pimmi

# TODO parameters from command line, fuse with generate_cluster_viz
logger = logging.getLogger("fuse ")

results_pattern = "index/dataset1.ivf1024.mining_0*.csv"
all_results_file = "index/dataset1.ivf1024.mining.all"


if __name__ == '__main__':
    pd.set_option('display.max_rows', 500)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.width', 1000)
    pd.set_option('display.max_colwidth', 1000)

all_similarities = []
for file in sorted(glob.glob(results_pattern)):
    logger.info(file)
    similarities = pd.read_csv(file, quoting=csv.QUOTE_NONNUMERIC)
    logger.info(" ~ similarity file loaded  : " + str(len(similarities)) + " edges")
    similarities = similarities.query('query_image_id != result_image_id and nb_match_ransac >= 10')
    logger.info(" ~ similarity file filtered: " + str(len(similarities)) + " edges")
    all_similarities.append(similarities)
logger.info("Concatenating results from " + str(len(all_similarities)) + " files")
all_similarities = pd.concat(all_similarities)
all_similarities[prm.dff_result_path] = "/" + all_similarities[prm.dff_result_path]
all_similarities[prm.dff_query_path] = "/" + all_similarities[prm.dff_query_path]
# all_similarities = all_similarities.astype({prm.dff_query_image: int, prm.dff_result_image: int})
with open(all_results_file, 'wb') as f:
    pickle.dump(all_similarities, f, pickle.HIGHEST_PROTOCOL)
f.close()

# logger.info("Loading results from " + all_results_file)
# with open(all_results_file, 'rb') as f:
#     all_similarities = pickle.load(f)
# f.close()
# logger.info(" ~ similarity file : " + str(len(all_similarities)) + " edges")
# print(all_similarities.head(20))